/********************************************************************************************************
*						MyRapBook v.0.1										                            *
*																		                                *
*				Copyright (C) 2008  Dimitris Tassopoulos								                *
*																		                                *
*	MyRapBook is a rapidshare download daemon for Western Digital's My Book World external hard drives.	*
*	It comes with a web interface and it's based on libcurl .									        *
*																		                                *
*	This program is free software: you can redistribute it and/or modify						        *
*	it under the terms of the GNU General Public License as published by						        *
*	the Free Software Foundation, either version 3 of the License, or							        *
*	any later version.															                        *
*																		                                *
*	This program is distributed in the hope that it will be useful,								        *
*	but WITHOUT ANY WARRANTY; without even the implied warranty of						                *
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the					                    *
*	GNU General Public License for more details.										                *
*																		                                *
*	You should have received a copy of the GNU General Public License							        *
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.							    *
*																		                                *
*-------------------------------------------------------------------------------------------------------*
*																		                                *
*	Details: 																                            *
*	1. This daemon is based on CURL library (http://curl.haxx.se/). Is running multiple threads and 	*
*	    it is able to download up to 4 files simultaneously.										    *
*	2. To install the daemon properly you have to follow the instructions in the READ.ME file or go to	*
*	    myrapbook.sourceforge.net web page to view the instructions there.						        *
*	3. Any news, bugs, information and new releases will be post to myrapbook.sourceforge.net			*
*	4. To make daemon to work properly you need the following packages installed:					    *
*		- web server (like apache, boa, lighttp etc.)										            *
*		- PHP 4 - 5 interpreter													                        *
*		- OpenSSL															                            *
*		- Curl with libcurl														                        *
*	    To install these packeges in WD's My Book World view instructions in my web site 			   	*
*	   (myrapbook.sourceforge.net).												                        *
*																		                                *
*********************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <fcntl.h>
#include <syslog.h>

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif


#define MAX_THREADS 5

#define MAX_ACCEPT_QUEUE 1  /* number of clients the server can queue. here one and only... YOU! */


/**** TYPE DEFINITIONS ****/

typedef struct {

    double file_size;       /*The size of file that was downloaded*/
    double speed;           /*Downloaded speed in Bytes/sec*/
    double conn_time;       /*Download duration*/
} tp_downloadInfo;



typedef struct {
    double dltotal;     /*Total file size to download*/
    double dlnow;       /*Downloaded so far*/
    double ultotal;     /*For upload. Just Ignore*/
    double ulnow;       /*For upload. Just Ignore*/
} tp_progressInfo;

typedef struct {
    char filename[250];
    int thread_id;
} tp_fileInfo;

struct download_hdlr {
    struct download_hdlr *   next;           /* Link field for linked list. */
    tp_fileInfo                 fileInfo;
    tp_progressInfo             progressInfo;
    tp_downloadInfo             downloadInfo;
} ;


/**** END OF TYPE DEFINITIONS ****/

/**** GLOBAL VARIABLES ****/

int                     DEBUG = 0;
struct download_hdlr*   download_queue;                                          /* A linked list of pending jobs. */
pthread_mutex_t         download_queue_mutex = PTHREAD_MUTEX_INITIALIZER;        /* A mutex protecting job_queue. */
sem_t                   download_queue_count;                                    /* A semaphore counting the number of jobs in the queue. */
int                     glb_max_threads = 4;
int                     glb_server_port = 7700;
char                    glb_log_file_name[] = "./deflog.txt";
char                    glb_downloadFolder[200];
char                    glb_wwwFolder[200];
char                    glb_downloadHistory[200];
int                     glb_client_is_connected = 0;
pthread_t               curl_thread[MAX_THREADS];
pthread_t               GOD;
char                    CONFIG_FILE[] = "./myrapbook.conf";
int                     glb_log_flag_available = 0;
int                     glb_history_flag_available = 0;
char					glb_user[25];
char					glb_pass[25];
struct download_hdlr*   glb_downloadStatus[4];


/**** END OF GLOBAL VARIABLES ****/

/**** FUNCTIONS ****/

int add_log(char * msg);
int add_in_history(char * filename);
int read_config_var( char * config_filename, char * param_to_find , char * value );
int check_valid_url(char * url);
int progress_callback(void *inc_download_handler, double dltotal, double dlnow, double ultotal, double ulnow);
size_t write_to_tmp_file(void *ptr, size_t size, size_t nmemb, FILE *stream);
size_t read_from_tmp_file(void *ptr, size_t size, size_t nmemb, FILE *stream);
int check_valid_url(char * url);
int get_filename_from_url(char * url, char * filename);
int create_tmp_filename(char * realname, char * temp_filename);
void initialize_download_queue ();
void enqueue_download (char * tdfilename);
void * start_downloading(void * thread_id);
int process_client_data(char * in_buffer, int datalen, int client_fd, int server_fd);
void * create_GOD(void *god_rules);
int get_parameter(int param, char* inStr, char* outStr);
int str_Search( char *strToFind, char *strToSearch );



/* C O D E */
int add_log(char * msg)
{
    time_t      now_time;           /* time type */
    struct tm * tp_now_time;    /* structure conatin year, month, -> , sec */
    FILE *      log_file;
    char        tmp_str[30];

    /* get timestamp */
    time(&now_time);

    /* convert time stamp to current local time */
    tp_now_time = localtime(&now_time);

    /* format data */
    strcpy(tmp_str, asctime(tp_now_time));

    /* search and destroy \n that asctime returns */
    tmp_str[strcspn ( tmp_str, "\n" )] = '\0';

    /* wait until the log file is available */
    while(!glb_log_flag_available);

    glb_log_flag_available = 0;

    /* open file for writing */
    log_file = fopen(glb_log_file_name, "a");

    /* save to file */
    if (log_file != NULL)
    {
        fprintf(log_file, "%s : %s\n", tmp_str, msg);
        if (DEBUG)
            fprintf(stderr, "%s\n", msg);

        /* close log */
        fclose(log_file);

        glb_log_flag_available = 1;
    }
    else {
        glb_log_flag_available = 1;
        return -1;
    }

    return 0;

}

/* Add downloaded file in history
* @param    : filename
* @return   : 0, if OK, else -1
*/
int add_in_history(char * filename)
{
    time_t      now_time;           /* time type */
    struct tm * tp_now_time;    /* structure conatin year, month, -> , sec */
    FILE *      history_file;
    char        tmp_str[30];

    /* get timestamp */
    time(&now_time);

    /* convert time stamp to current local time */
    tp_now_time = localtime(&now_time);

    /* format data */
    strcpy(tmp_str, asctime(tp_now_time));

    /* search and destroy \n that asctime returns */
    tmp_str[strcspn ( tmp_str, "\n" )] = '\0';

    /* wait until the log file is available */
    while(!glb_history_flag_available);

    glb_history_flag_available = 0;

    /* open file for writing */
    history_file = fopen(glb_downloadHistory, "a");

    /* save to file */
    if (history_file != NULL)
    {
        fprintf(history_file, "%s : %s\n", filename, tmp_str);
        if (DEBUG)
            fprintf(stderr, "Added to history: %s\n", filename);

        /* close log */
        fclose(history_file);

        glb_history_flag_available = 1;
    }
    else {
        glb_history_flag_available = 1;
        return -1;
    }

    return 0;
}

int copy_file(char * source, char * target)
{

    int rc = 0;
    FILE *fpin = fopen(source, "r");

    if (DEBUG)
        fprintf(stderr, "copy %s -> %s\n", source, target);

    if(fpin != NULL)
    {
        FILE *fpout = fopen(target, "w");
        if(fpout != NULL)
        {
            int ch = 0;
            while((ch = getc(fpin)) != EOF)
            {
                putc(ch, fpout);
            }
            if(ferror(fpin))
            {
                rc |= 1; //COPYFILE_READ_ERROR;
            }
            if(ferror(fpout))
            {
                rc |= 2; //COPYFILE_WRITE_ERROR;
            }
            if(fclose(fpout) != 0)
            {
                rc |= 4; //COPYFILE_OUTCLOSE_ERROR;
            }
        }
        else
        {
            rc |= 8; //COPYFILE_OUTOPEN_ERROR;
        }

        if(fclose(fpin) != 0)
        {
            rc |= 16; //COPYFILE_INCLOSE_ERROR;
        }
    }
    else
    {
        rc |= 32; //COPYFILE_INOPEN_ERROR;
    }

    return rc;

}

/**
 * Search in string
 * @param strToFind
 * @param strToSearch
 *
 * @return s8
 */
int str_Search( char *strToFind, char *strToSearch )
{
    int i,k,l,len,strLen;

    len = strlen(strToFind);         // finds the string length of 'StrToFind'
    strLen = strlen(strToSearch);    // finds the string length of 'strToSearch'

    for (i=0; i<strLen; i++)
    {
        if (strToFind[0] == strToSearch[i]) // Searches in the 'strToSearch' for the beginning of 'strToFind'
        {
            k = i;
            for (l=0; l<len;l++)
            {
                if (strToFind[l] == strToSearch[k]) // compares the two strings
                    k++;
                else
                    break;
            }
            if (len == l)
            {
                return(i);  // Returns the position of the beginning of the string
            }
        }
    }
    return(-1);
}

/**
 * Get Parameter from a string
 * @param param
 * @param inStr
 * @param outStr
 *
 * @return int parameter length
 */
int get_parameter( int param, char * inStr, char* outStr)
{
    int i,len ,com ,equ, com2, parlen;

    len = strlen( inStr );
    equ = str_Search( "=", inStr ) ;

    if ( len <= equ || equ == -1) goto errLabel ;

    if ( param == 1) {
        com =  str_Search( ",", inStr ) ;
        if (com == -1) { // only one parameter ?
            parlen = len - equ - 2;
            memcpy( outStr, inStr + equ + 1 , parlen );
            outStr[ parlen ] = 0;
        } else {        // comma at the end of parameter
            parlen = com - equ - 1 ;
            memcpy( outStr, inStr + equ + 1 , parlen );
            outStr[ parlen ] = 0;
        }
    }  else {       // not the first parameter
        com=0;
        for ( i=1 ; i<param ; i++) {    // find the comma of interest
            com2 = str_Search( ",", inStr + com + 1 ) ;
            if ( com2 == -1) goto errLabel ;
            com = com2 + com + 1;
        }

        com2 =  str_Search( ",", inStr + com + 1 ) ;

        if (com2 == -1) { // only one comma ?
            parlen = len - com - 2  ;
            //TRACE((4, "Parlen : %d", parlen ));
            memcpy( outStr, inStr + com + 1 , parlen );
            outStr[ parlen ] = 0;
        } else {
            parlen = com2 ;
            memcpy( outStr, inStr + com + 1 , parlen );
            outStr[ parlen ] = 0;
        }
    }

    return parlen;

errLabel:
    return -1;
}

/* Get a parameter from config file
* This function is used to extract parameters from the default config file.
* The defau;t config file is: "myrapbook.conf"
*/
int read_config_var( char * config_filename, char * param_to_find , char * value )
{
    int     len;
    FILE *  fd;
    char    tmp_str[80];

	if( param_to_find == NULL ) return (-1);

	len = strlen(param_to_find);


	if( len > 77) return(-2);


	if(strlen(config_filename) > 0 )
	{
		fd = fopen(config_filename, "r");
		if (fd == NULL) return(-2);
	}
	else
		return(-3);


	if( fseek(fd, 0, SEEK_SET) )
	{
		fclose(fd);
		return(-1);
	}


	for(;;)
	{
		fgets(tmp_str, 80, fd);

		if( ferror(fd) || feof(fd) ) return(-2);

		len = strlen(tmp_str);

		if( strncmp(param_to_find, tmp_str, strlen(param_to_find)) == 0 )
		{
				if (tmp_str[len - 1] == '\n') tmp_str[--len] = 0;
                    sprintf(value, "%s", &tmp_str[strlen(param_to_find)+1] );
                break;
		}
	}

	fclose(fd);

return 0;

}  /* end of read_config_var */

/* Progress callback handler
* Curl automatically calls this handler and refresh the total and current download size.
* These parameters are stored in the download handler.
*
* @param: inc_download_handler, the incoming download handler
* @param: dltotal, the total size of the file in bytes
* @param: dlnow, the downloaded size in bytes
* @param: ultotal, the total upload file size in bytes, not used here
* @param: ulnow, the uploaded size in bytes, not used here
* @return: always 0
*/
int progress_callback(void *inc_download_handler, double dltotal, double dlnow, double ultotal, double ulnow)
{
    int k;
    /* create a temp handler */
    struct download_hdlr* tmp_download_hdlr;

    /* grab the incoming handler */
    tmp_download_hdlr = (struct download_hdlr*) inc_download_handler;

    /* retrieve total and downloaded file size */
    tmp_download_hdlr->progressInfo.dltotal = dltotal;
    tmp_download_hdlr->progressInfo.dlnow = dlnow;


    if (DEBUG) {
        /* print for debugging*/
       fprintf(stderr, "\rFile (%d): %d KB/ %d KB (%4.2g %%)",
                (int) (tmp_download_hdlr->fileInfo.thread_id),
                (int) (tmp_download_hdlr->progressInfo.dlnow/1024), (int) (tmp_download_hdlr->progressInfo.dltotal/1024),
                tmp_download_hdlr->progressInfo.dlnow * 100.0/tmp_download_hdlr->progressInfo.dltotal);
    }

    k = (int) (tmp_download_hdlr->fileInfo.thread_id);
    k -= 1;
    glb_downloadStatus[k] = tmp_download_hdlr;
    if (DEBUG)
        fprintf(stderr, "   %d", k);

    /* the temp handler will now suicide but the changes in incoming handler will remain */
    return 0;
}

/* Write to file stream
* This is a callback function called from curl and writes the downloaded data
* to the input buffer stream/file.
*
*/
size_t write_to_tmp_file(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
    return fwrite(ptr, size, nmemb, stream);
}

/* Read from file stream
* This is a callback function called from curl and reads the downloaded data
* to the input buffer stream/file.
*
*/
size_t read_from_tmp_file(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
    return fread(ptr, size, nmemb, stream);
}


/* Check if given url is valid
* This function checks if a given url is a valid rapidshare url. It only
* checks url length and prefix, therefore is not very secure.
*
* @param: url string
* @return:  0, if url seemd to be valid
*          -1, if the url is too short to be valid
*          -2, if the filename prefix is not valid
*/
int check_valid_url(char * url)
{
    char log_buf[250];
    /* check if string length is lower than 38. (http://rapidshare.com/files/) = 38 chars */
    if ( strlen(url) < 38 ) {
        return -1;
    }

    /* check if filename contains the standard string (http://rapidshare.com/files/) */
    if ( strncmp(url, "http://rapidshare.com/files/", 28) == 0) {
        return 0;
    } else {
        return -2;
    }
}

/* Retrieve filename from url
* This function extracts the real filename of the given url that will
* be used to create the temp and result filename.
*
* @param:   url string
* @param:   resulting real filename
* @return:  0, if there's no error
*          -1, if url was too short
*          -2, if url was not a valid rapidshare
*/
int get_filename_from_url(char * url, char * filename)
{
    char    log_buf[250];
    int     i = 0;
    int     resp = 0;               /* hi I'm a response var. I have a great family and a lot of brothers in this code */

    /* run snippet for valid url string */
    resp = check_valid_url(url);

    if (resp == -1) {
        /* Oops... errrrrror */
        snprintf(log_buf, 250, "URL (%s) is too short!\n", url);
        add_log(log_buf);
        return resp;
    } else if (resp == -2) {
        /* Oops... error again */
        snprintf(log_buf, 250, "Not valid rapidshare link: %s\n", url);
        add_log(log_buf);
        return resp;
    }

    /* 37 is always the position of the first character of the real filename */
    for (i=37; i < strlen(url); i++)
        *(filename + i - 37) = *(url + i);  /* bring it to me */

    *(filename + i - 37) = '\0';    /* don't you ever forget to terminate those damn strings cause they don't */

    return 0;
}

/* Create a temp from the real filename
* This function creates a temp filename that is similar to the real filename.
* (It just add a ".temp" fix in the end of real name. The download data are
* stored in the temp file and only after the downloading finish without errors,
* the filename is renamed to the original filename.
*
* @param:   real name
* @param:   resulting temp filename
* @return:  the length of filename, if there's no error
*          -1, if name creation fails
*/
int create_tmp_filename(char * realname, char * temp_filename)
{
    char    log_buf[250];
    int     resp;   /* hi have you seen my brothers responses var? */
    int     filelen = strlen(realname);

    /* add a fix name (.temp) to the real name */
    resp = snprintf(temp_filename, filelen + 6, "%s.temp", realname);

    if (resp > 0)
        return resp;    /* nice */
    else {
        snprintf(log_buf, 250, "Error creating temp filename.");
        add_log(log_buf);
        return -1;  /* geeez */
    }
}

/* Initialize download queue
* Everything needs some initialization, also the download queue list
*/
void initialize_download_queue ()
{
    /* The queue is initially empty. */
    download_queue = NULL;

    /* Initialize the semaphore which counts jobs in the queue. Itsinitial value should be zero. */
    sem_init (&download_queue_count, 0, 0);
}

/* Add a download to the queue
* By calling this function a new download request is added to the top of the
* existing download queue list.
*
* @param: tdfilename, the filename to download
*/
void enqueue_download (char * tdfilename)
{
    struct download_hdlr* new_download;

    /* Allocate a new job object. */
    new_download = (struct download_hdlr*) malloc (sizeof (struct download_hdlr));

    /* Set the other fields of the download struct here... */
    /* Lock the mutex on the download queue before accessing it. */
    pthread_mutex_lock (&download_queue_mutex);

    strcpy(new_download->fileInfo.filename, tdfilename);

    /* Place the new download at the head of the queue. */
    new_download->next = download_queue;
    download_queue = new_download;

    /* Post to the semaphore to indicate that another download is available. If
    threads are blocked, waiting on the semaphore, one will become
    unblocked so it can process the download. */
    sem_post (&download_queue_count);

    /* Unlock the job queue mutex. */
    pthread_mutex_unlock (&download_queue_mutex);
}


void * start_downloading(void * thread_id)
{
    char        log_buf[550];
    CURL *      curl_hdlr;            /* get a curl handler */
    CURLcode    curl_resp;         /* hey i'm a curl resp handler, have nothing to do with the other lazy resp vars*/
    FILE *      fd;                   /* file descriptor for saving the downloaded data */
    char        tmpFilename[200];      /* keeps the temp filename. can be smaller... */
    char        realname[200];         /* keeps the real filename */
    int         resp = 0;               /* you know what i am */
    int 		thread_err = 1;
    char        realnameWithPath[400];
    char        tmpFilenameWithPath[400];


    /* thread started normally*/
    //snprintf(log_buf, 250, "Thread %d started.", (int) thread_id);
    //add_log(log_buf);

    while (1) {

        struct download_hdlr* next_download;
        /* Wait on the job queue semaphore. If its value is positive,
        indicating that the queue is not empty, decrement the count by
        1. If the queue is empty, block until a new job is enqueued. */
        sem_wait (&download_queue_count);
        /* Lock the mutex on the job queue. */
        pthread_mutex_lock (&download_queue_mutex);
        /* Because of the semaphore, we know the queue is not empty. Get
        the next available job. */
        next_download = download_queue;
        /* Remove this job from the list. */
        download_queue = download_queue->next;
        /* Unlock the mutex on the job queue because we’re done with the
        queue for now. */
        pthread_mutex_unlock (&download_queue_mutex);

        /* check if other threads are alive */


        /* initialization needed here */
        curl_hdlr = curl_easy_init();

        if(curl_hdlr) { /* ok we got it init */

            /* get thread id */
            next_download->fileInfo.thread_id = (int) thread_id;
            if (DEBUG)
                fprintf(stderr, "My new thead id is: %d.\n", next_download->fileInfo.thread_id);

            /* get filename */
            if (!(get_filename_from_url(next_download->fileInfo.filename, realname))) {
                if (DEBUG)
                 fprintf(stderr, "Real file name is: %s\n", realname);
            }
            else {
                /* log error and return */
                free (next_download);
                pthread_exit((void*) thread_err);
            }

            /*create temp file to write data*/
            if (create_tmp_filename(realname, tmpFilename) > 0) {
                if (DEBUG)
                    fprintf(stderr, "Temp filename: %s\n", tmpFilename);
            }
            else {
                /* log error and return */
                free (next_download);
                pthread_exit((void*)thread_err);
            }

            /* append download folder path to file that curl will use */
            strcpy(tmpFilenameWithPath, glb_downloadFolder);
            strcat(tmpFilenameWithPath, tmpFilename);
            fd = fopen(tmpFilenameWithPath, "w");

            if (fd < 0) {
                snprintf(log_buf, 550, "Error while opening temp file (%d)): %s", (int)thread_id, tmpFilenameWithPath);
                add_log(log_buf);
                free (next_download);
                pthread_exit((void*) thread_err);
            }

            /*Get file to download*/
            curl_easy_setopt(curl_hdlr, CURLOPT_URL, next_download->fileInfo.filename);

            /*Get cookie from rapidshare to use*/
            curl_easy_setopt(curl_hdlr, CURLOPT_COOKIEFILE, "./cookie");

            /*Set curl to follow all the links that rapidshare server provides untill reach download file*/
            curl_easy_setopt(curl_hdlr, CURLOPT_FOLLOWLOCATION, 1);

            /*Store incoming data to file*/
            curl_easy_setopt(curl_hdlr, CURLOPT_WRITEDATA, fd);

            /*Set calling function. Now my_write_func will use as FILE * stream the file given in CURLOPT_WRITEDATA*/
            curl_easy_setopt(curl_hdlr, CURLOPT_WRITEFUNCTION, write_to_tmp_file);

            /*Read function*/
            curl_easy_setopt(curl_hdlr, CURLOPT_READFUNCTION, read_from_tmp_file);

            /*User agent for MyBook downloader*/
            curl_easy_setopt(curl_hdlr, CURLOPT_USERAGENT, "Jaco-MyRapidBook/1.0");

            /*Enable progress watch*/
            curl_easy_setopt(curl_hdlr, CURLOPT_NOPROGRESS, FALSE);

            curl_easy_setopt(curl_hdlr, CURLOPT_PROGRESSFUNCTION, progress_callback);

            curl_easy_setopt(curl_hdlr, CURLOPT_PROGRESSDATA, next_download);

            snprintf(log_buf, 550, "Start downloading (%d): %s.", (int) thread_id, tmpFilename);
            add_log(log_buf);

            /*Start downloading*/
            curl_resp = curl_easy_perform(curl_hdlr);

            /*Close file*/
            fclose(fd);

            snprintf(log_buf, 550, "Download finished (%d): %s.\nwith resp: %d\n", (int) thread_id, tmpFilename, (int)curl_resp);
            add_log(log_buf);


            /*Rename temp file to real filename*/

            strcpy(realnameWithPath, glb_downloadFolder);
            strcat(realnameWithPath, realname);

            resp = rename(tmpFilenameWithPath, realnameWithPath);
            if (resp != 0) {
                /* log error and return */
                snprintf(log_buf, 550, "Error renaming file from (%s) to (%s).", tmpFilenameWithPath, realnameWithPath);
                add_log(log_buf);
                free (next_download);
                pthread_exit((void*) thread_err);
            }



            if (curl_resp == CURLE_OK) {

                /* add file in history */
                add_in_history(realname);

                /* retrieve some stats for the download */

                /* get downloaded file size */
                curl_resp = curl_easy_getinfo(curl_hdlr, CURLINFO_SIZE_DOWNLOAD, &next_download->downloadInfo.file_size);
                if((curl_resp == CURLE_OK) && next_download->downloadInfo.file_size) /* in bytes */
                    next_download->downloadInfo.file_size /= 1024;   /* convert size to KB */

                /* get connection time */
                curl_easy_getinfo(curl_hdlr, CURLINFO_TOTAL_TIME, &next_download->downloadInfo.conn_time);
                if((curl_resp == CURLE_OK) && next_download->downloadInfo.conn_time)  /* in seconds */

                /* get download average speed */
                curl_easy_getinfo(curl_hdlr, CURLINFO_SPEED_DOWNLOAD, &next_download->downloadInfo.speed);
                if((curl_resp == CURLE_OK) && next_download->downloadInfo.speed)  /* in bps */
                    next_download->downloadInfo.speed /= 1024;  /* convert to kbps */


                snprintf(log_buf, 550, "Stats:\n"
                                    "\tFilename: %s\n"
                                    "\tDownloaded size: %g KB\n"
                                    "\tDownloaded time: %g\n"
                                    "\tDownloaded speed: %g Kbps.",
                                    realname,
                                    next_download->downloadInfo.file_size,
                                    next_download->downloadInfo.conn_time,
                                    next_download->downloadInfo.speed);



                add_log(log_buf);
                {
                    int kill_id = (int)thread_id;
                    glb_downloadStatus[kill_id -1] = NULL;
                }




            } else {    /* download failed */
                if (DEBUG)
                    fprintf(stderr, "An error (%d) occured.\n", curl_resp);
                /* log error and return */
                free (next_download);
                pthread_exit((void*) thread_err);
            }

            /* always cleanup */
            curl_easy_cleanup(curl_hdlr);


            /* we're done with libcurl, so clean it up */
            curl_global_cleanup();

       }

        /* Clean up. */
        free (next_download);
    }
}

/*!!!!!!! ALL INCOMING DATA ENDS WITH 0x13 and 0x14 !!!!!!*/

int process_client_data(char * in_buffer, int datalen, int client_fd, int server_fd)
{
    int     nbytes = 0;             /* number of bytes for above buffers */
    char    buffer[250];  /* 1K5 outcoming buffer */
    //int     tcp_resp = 0;           /* holds tcp responses */
    int     i = 0;
    char    out_buffer[1000];

    if (DEBUG) {
        memcpy(buffer, in_buffer, datalen);
        buffer[datalen] = 0;
        fprintf(stderr, "incoming string: %s\nincoming bytes: %d.\n", buffer, datalen);
    }


    if (!(strncmp(in_buffer, "AUTH=", 5))) {
    	char user[17];
    	char r_user[17];
    	char ip[25];
    	int *fd;

        /* get user name */
        if (get_parameter(1, in_buffer, user) < 1) {
    		 strcpy(out_buffer, "VER_FAILED");
    		 add_log("user not found.");
    		 if (DEBUG)
                fprintf(stderr, "conf user: %s\nuser's ip: %s\n", user, ip);
    	}


        /* get user's ip */
        if (get_parameter(2, in_buffer, ip) < 1) {
    		 strcpy(out_buffer, "VER_FAILED");
    		 add_log("ip not found.");
    		 if (DEBUG)
                fprintf(stderr, "conf user: %s\nuser's ip: %s\n", user, ip);
    	}

    	read_config_var("./myrapbook.conf", "USER", r_user);
    	if (!(strcmp(user, r_user))) {
    	    FILE *fd;
            strcpy(out_buffer, glb_pass);
            if (DEBUG)
                fprintf(stderr, "user: %s\nip: %s\n", user, ip);
            add_log("valid user.");
            /* now save in a file the user that is connected with his ip */
            fd = fopen("./tmp.usr", "w");
            /* save to file */
            if (fd != NULL)
            {
                time_t timestamp;
                time(&timestamp);
                fprintf(fd, "%s,%s,%li", user, ip, timestamp);
                if (DEBUG)
                    fprintf(stderr, "%s,%s,%li", user, ip, timestamp);

                /* close log */
                fclose(fd);
            }
            else {
                strcpy(out_buffer, "VER_FAILED");
                add_log("could not create user file.");
            }

    	}
    	else {
    		 strcpy(out_buffer, "VER_FAILED");
    		 add_log("invalid user.");
    	}

    }
    else if (!(strncmp(in_buffer, "DOWN=", 5))) {
        char url[120];
    	char ip[25];
    	int *fd;

        /* a link for downloading follows */
        /* retrieve link */

        /* get user's ip */
        if (get_parameter(1, in_buffer, ip) < 1) {
    		 strcpy(out_buffer, "DOWN_FAIL");
    		 add_log("ip not found.");
    	}


        /* get url */
        if (get_parameter(2, in_buffer, url) < 1) {
    		 strcpy(out_buffer, "DOWN_FAIL");
    		 add_log("url not found.");
    	}

    	enqueue_download(url);
        strcpy(out_buffer, "DOWN_OK");

        /*if (check_valid_connection(ip, user)) {


            // now append file to download list
            enqueue_download(buffer);
            strcpy(out_buffer, "Download enqued.\n");
        }
        else {
    		 strcpy(out_buffer, "TIMED_OUT");
    		 add_log("connection timed out.");
    	}*/
    }
    else if (!(strncmp(in_buffer, "DSTATUS", 7))) {
        char tmpStr[250];
        out_buffer[0] = 0;
        if (DEBUG)
            fprintf(stderr, "\nDSTATUS detected\n");
        for ( i = 0; i<4; i++) {
            if ( glb_downloadStatus[i] != NULL ) {
                if (DEBUG)
                    fprintf(stderr, "will send: %s\n", glb_downloadStatus[i]->fileInfo.filename);
                snprintf(tmpStr, 250, "DSTATUS=%d,%s,%d,%d,\n",
                            glb_downloadStatus[i]->fileInfo.thread_id,
                            glb_downloadStatus[i]->fileInfo.filename,
                            (int)glb_downloadStatus[i]->progressInfo.dltotal/1024,
                            (int)glb_downloadStatus[i]->progressInfo.dlnow/1024);
                strcat(out_buffer, tmpStr);
            }
            else {
                snprintf(tmpStr, 250, "DSTATUS=%d,0,0,0,\n", i+1);
                strcat(out_buffer, tmpStr);
            }
        }

    }
    else if (!(strncmp(in_buffer, "GETHISTORY", 10))) {
        /* send history file to client */
        int resp;
        char tmp_filename[200];

        /* this will create a temp file in www folder and will copy
        the contents of the history file in there. then the php will
        retrieve the file and after finish with it will delete it.*/

        strcpy(tmp_filename, glb_wwwFolder);
        strcat(tmp_filename, "tmphist.txt");

        //strcpy(tmp_filename, "./tmphist.txt");

        resp = copy_file(glb_downloadHistory, tmp_filename);
        if (DEBUG)
            fprintf(stderr, "Copy filename result: %d\n", resp);

        if (!resp)
            strcpy(out_buffer, "GETHISTORY_OK");
        else
            strcpy(out_buffer, "GETHISTORY_FAIL");


    }
    else if (!(strncmp(in_buffer, "KILLTMPHIST", 11))) {
        /* kill tmphist.txt */
        char tmp_filename[200];
        FILE * tmp_fd;

        strcpy(tmp_filename, glb_wwwFolder);
        strcat(tmp_filename, "tmphist.txt");

        /* open tmp file */
        tmp_fd = fopen(tmp_filename, "w");
        if (tmp_fd != NULL) {
            fclose (tmp_fd); //now tmphist.txt is empty
            /* kill it */
            remove(tmp_filename);
            strcpy(out_buffer, "KILLTMPHIST_OK");
        }
        else {
            strcpy(out_buffer, "KILLTMPHIST_FAIL");
        }

    }
    else if (!(strncmp(in_buffer, "CLRHISTORY", 10))) {
        /* Clear History File */
        FILE * fd;
        /* open file for writing */
        fd = fopen(glb_downloadHistory, "w");

        /* save to file */
        if (fd != NULL)
        {
            /* close log */
            fclose(fd);
            strcpy(out_buffer, "CLRHISTORY_OK");
        }
        else {
            strcpy(out_buffer, "CLRHISTORY_FAIL");
        }

    }
    else if (!(strncmp(in_buffer, "THREADS", 7))) {
        for ( i = 1; i <= glb_max_threads; i++ )
            if (DEBUG)
                fprintf(stderr, "Thread= %d\n", i);

    } else {
        strcpy(out_buffer, "What? I can't understand what you're saying!\n");
    }


    /* send response */
    nbytes = strlen(out_buffer);

    if ( nbytes > 0 ) {
        if(send(client_fd, out_buffer, nbytes, 0) == -1) { /* maybe client is bored and gone */
            if (DEBUG)
                fprintf(stderr, "send() error.\n");
            return -1;
        }
        if (DEBUG)
            fprintf(stderr, "Sended response: %s", out_buffer);
    } else {
        if (DEBUG)
            fprintf(stderr, "Sending response error.\n");
        return -2;

    }

    return 0;
}


/* The threads God
* If a thread suddenly dies under any circumstances then the god take cares of its resurrection
* God supposed will never die as long the program exists.
*/
void * create_GOD(void *god_rules)
{
    int i = 0;
    int resp;
    int status;
    char log_buf[200];

    if (DEBUG)
        fprintf(stderr, "God created.\n");

    while(1)
    {
        /*Create threads*/
        for (i=1; i <= glb_max_threads; i++) {
            /* start thread */
            resp = pthread_create(&curl_thread[i], NULL, start_downloading, (int *) i);

            if (resp !=0 ){
                /* it's fucked up */
                snprintf(log_buf, 200, "Couldn't run thread, errno %d\n", resp);
                add_log(log_buf);
            }
        }

        /* wait for unlucky threads that will die */
        while(1) {
            for (i=1; i<= glb_max_threads; i++)
            {
                pthread_join (curl_thread[i], (void*) &status);
                /* welcome to threads' paradise */

                /* time to go now*/
                resp = pthread_create(&curl_thread[i], NULL, start_downloading, (int *) i);

                if (resp !=0 ){
                    /* it's fucked up */
                    snprintf(log_buf, 200, "Couldn't run thread, errno %d\n", resp);
                    add_log(log_buf);
                }
            }
        }
    }
}





/**** END OF FUNCTIONS ****/


/* Main Loop
*
*/
int main(int argc, char * argv[])
{
    int     i = 0;
    char    tcp_buffer_in[1500];   /* 1K5 incoming buffer */
    int     nbytes;                 /* number of bytes for above buffers */
    int     tcp_resp;               /* holds tcp responses */
    char    log_buf[250];
    pid_t   pid;
    pid_t   sid;

    /* Fork off the parent process */
    pid = fork();

    if (pid < 0) {
        exit(EXIT_FAILURE);
    }

    /* Exit parent process */
    if (pid > 0) {
        exit(EXIT_SUCCESS);
    }

    /* Change the file mode mask */
    umask(0);

    /* Create a new SID for the child process */
    sid = setsid();
    if (sid < 0 ) {
        add_log("Fail to create new sid.");
        exit(EXIT_FAILURE);
    }

    /* Change dir to /tmp/myrapbook */
    if ((chdir("/tmp/myrapbook")) < 0) {
        exit(EXIT_FAILURE);
    }

	/* If argument is --DEBUG then display informations */
	if (argv[1] != NULL) {
        fprintf(stderr, "argv: %s", argv[1]);
        if (!(strncmp(argv[1], "--DEBUG", 7))) {
            DEBUG = TRUE;
        }
        else {
            DEBUG = FALSE;
        }
    }

    if (DEBUG)
        fprintf(stderr, "\n\tDEBUG MODE IS ON.\n\n");

    /*** retrieve parameters from config file. If they're not found then global deafult are used. ***/

    /* get log path */
    if (!(read_config_var("./myrapbook.conf", "LogFileName", tcp_buffer_in))) {
        strcpy(glb_log_file_name, tcp_buffer_in);
        if (DEBUG)
            fprintf(stderr, "Found in config LogFilePath: %s\n", glb_log_file_name);
    }

    /* flag log is available for write */
    glb_log_flag_available = 1;

    add_log("\n\t\tDaemon started.");

    /* Adding sid to log makes it easier to kill the daemon */
    snprintf(log_buf, 250, "pid: %ld, sid: %ld", (long int)pid, (long int)sid);
    add_log(log_buf);

    /* Close standard file descriptors */
    if (!DEBUG) {
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);
    }

    /* get server port*/
    if (!(read_config_var("./myrapbook.conf", "port", tcp_buffer_in))) {
        if ( (tcp_resp = atoi(tcp_buffer_in)) > 0 ) {
            glb_server_port = tcp_resp;
            snprintf(log_buf, 250, "Port: %d", glb_server_port);
            add_log(log_buf);
        }
        else
            add_log("Error in port conversion. keeping deafult = 7700.");
    }

    /* get max connections / threads */
    if (!(read_config_var("./myrapbook.conf", "MAX_THREADS", tcp_buffer_in))) {

        if ( (tcp_resp = atoi(tcp_buffer_in)) > 0 ) {
            glb_max_threads = tcp_resp;
            snprintf(log_buf, 250, "MAX_THREADS: %d", glb_max_threads);
            add_log(log_buf);
        }
        else
            add_log("Error in MAX_THREADS conversion. keeping default = 4.");
    }

    /* get user name */
    if (!(read_config_var("./myrapbook.conf", "USER", tcp_buffer_in))) {
         strcpy(glb_user, tcp_buffer_in);
         snprintf(log_buf, 250, "user: %s", glb_user);
         add_log(log_buf);
    }
	else {
         add_log("No USER found! Exiting.");
         return 0;
    }

    /* get max connections / threads */
    if (!(read_config_var("./myrapbook.conf", "PASS", tcp_buffer_in))) {
         strcpy(glb_pass, tcp_buffer_in);
         snprintf(log_buf, 250, "pass: %s", glb_pass);
         add_log(log_buf);
    }
	else {
         add_log("No PASS for USER found! Exiting.");
         return 0;
    }

    /* get downloads folder */
    if (!(read_config_var("./myrapbook.conf", "DownloadFolder", tcp_buffer_in))) {
         strcpy(glb_downloadFolder, tcp_buffer_in);
         snprintf(log_buf, 250, "DownloadFolder: %s", glb_downloadFolder);
         add_log(log_buf);
    }
	else {
         add_log("No DownloadFolder found! Exiting.");
         return 0;
    }

    /* get downloads history file */
    if (!(read_config_var("./myrapbook.conf", "DownloadHistory", tcp_buffer_in))) {
         strcpy(glb_downloadHistory, tcp_buffer_in);
         snprintf(log_buf, 250, "DownloadHistory: %s", glb_downloadHistory);
         add_log(log_buf);
         glb_history_flag_available = 1;
    }
	else {
         add_log("No DownloadHistory found! Exiting.");
         return 0;
    }

    /* get WWWFolder */
    if (!(read_config_var("./myrapbook.conf", "WWWFolder", tcp_buffer_in))) {
         strcpy(glb_wwwFolder, tcp_buffer_in);
         snprintf(log_buf, 250, "WWWFolder: %s", glb_wwwFolder);
         add_log(log_buf);
    }
	else {
         add_log("No WWWFolder found! Exiting.");
         return 0;
    }

    /* create GOD (although that sounds weird) */
    tcp_resp = pthread_create(&GOD, NULL, create_GOD, "THOU SHALL RULE");

    /*** Init glb_downloadStatus ***/
    for (i=0; i<3; i++)
        glb_downloadStatus[i] = 0;


    /*** Initialize Server for accepting connections ***/

    /* server file descriptor */
    int server_fd;

    /* client file descriptor */
    int client_fd;

    /* local server address information */
    struct sockaddr_in server_addr;

    /* remote client address information */
    struct sockaddr_in client_addr;

    int client_addr_size;

    /*create server end point*/
    server_fd = socket(AF_INET, SOCK_STREAM, 0);
    if(server_fd == -1)
    {
        /*fail to create endpoint*/
        add_log("Failed to create socket. Terminate.");
        exit(1);
    }
    else
        /*succesfully created server endpoint*/
        add_log("Socket created succefully.");

    /* set server address family to inet */
    server_addr.sin_family = AF_INET;

    /* set server listening port */
    server_addr.sin_port = htons(glb_server_port);

    /* set server address to any incoming */
    server_addr.sin_addr.s_addr = INADDR_ANY;

    /* by default set sin_zero 8 bytes to 0x00 */
    memset(&(server_addr.sin_zero), 0, 8);

    /* bind initialized socket */
    if(bind(server_fd, (struct sockaddr *)&server_addr, sizeof(struct sockaddr)) == -1)
    {
        /* failed to bind socket*/
        add_log("Error binding socket.");
        /* retry to bind for ever */
        while ( bind(server_fd, (struct sockaddr *)&server_addr, sizeof(struct sockaddr)) < 0) {
                /* don't overhead. give it some time */
                sleep(5);   /* wait for 5 seconds */
                add_log("Retrying to bind.");
        };
        add_log("Got bind.");
    }
    else
        /* socket binded succesfully*/
        add_log("Bind.");

    /* start listening for incoming clients */
    if(listen(server_fd, MAX_ACCEPT_QUEUE) == -1)
    {
        /* failed to enter listening mode */
        add_log("Failed to enter in listening mode.");
        exit(1);
    }
    else
        /* it's alive and kicking*/
        add_log("Server 'sup.");

    while(1) {  /* loop for ever and accept incoming connections*/

        /* blocking code of listen */
        client_addr_size = sizeof(client_addr);
        if((client_fd = accept(server_fd, (struct sockaddr *)&client_addr, &client_addr_size)) == -1)
        {
            add_log("Server accept error.");
            exit(1);
        }
        else {
			/* This result in a huge logfile and cost performance while it's happening many times.
				If you want to log all incoming connections then remove comments. */
            /*
			snprintf(log_buf, 250, "%s: New connection on socket %d.", (char*) inet_ntoa(client_addr.sin_addr), client_fd);
			add_log(log_buf);
			*/
        }

        /* flag client connection */
        glb_client_is_connected = 1;

        while(glb_client_is_connected) {
            /* receive incoming data */
            if((nbytes = recv(client_fd, tcp_buffer_in, sizeof(tcp_buffer_in), 0)) <= 0)
            {
                /* got error or connection closed by client */
                if(nbytes == 0) {
                    /* connection closed */
					/*
					snprintf(log_buf, 250, "Socket %d hung up.", client_fd);
					add_log(log_buf);
					*/
                    close(client_fd);
                    glb_client_is_connected = 0;
                } else {
                    snprintf(log_buf, 250, "Error reading incoming data from %s.", (char*) inet_ntoa(client_addr.sin_addr));
                    add_log(log_buf);
                    /* close it... */
                    close(client_fd);
                    glb_client_is_connected = 0;
                }
            } else {
                /* process the incoming data */
                tcp_resp = process_client_data(tcp_buffer_in, nbytes, client_fd, server_fd);
                if (tcp_resp == -1)
                {
                    /* finished with client */
                    close(client_fd);
                    glb_client_is_connected = 0;
                }

            }
        }
    }
   return 0;
}
